/***

   4x DC ( RGB Led + DC MOTOR on/off )

 ***/
// MySensors SETUP
#define MY_RADIO_NRF24
#define MY_NODE_ID AUTO
#define MY_BAUD_RATE 115200
//#define MY_DEBUG    // Enables debug messages in the serial log

// PIN Definitions
#define RGBW
#define RED    3
#define GREEN  5
#define BLUE   6
#define MOTOR  7

// Includes
#include <MySensors.h>

#define NAME "myHitTrain"
#define RELEASE "1.3"

// Version history
// 1.0 26-nov-19: Copied from myDCDimmer 1.1

// RGB related-variables
enum ANIMATIOMODES {RAINBOW = 0, RANDONMIZE, FADERGB, FADEMULTICOLR, FLASHCOLOR, LAST_PROGRAM};

byte FADE_RGB_COLOR_MODES[]   = {0b0010, 0b0011, 0b0100, 0b0101, 0b1000, 0b1001, 0xFF};
byte FADE_MULTI_COLOR_MODES[] = {0b0010, 0b0011, 0b0110, 0b0111, 0b0100, 0b0101, 0b1100, 0b1101, 0b1000, 0b1001, 0b1010, 0b1011, 0xFF};

#ifdef RGBW
#define NODENAME "RGBW node"
#define NODETYPE S_RGBW_LIGHT
byte rgb_pins[] = {RED, GREEN, BLUE, MOTOR};
#else
#define NODENAME "RGB node"
#define NODETYPE S_RGB_LIGHT
byte rgb_pins[]   = {RED, GREEN, BLUE};
#endif

byte ledOffValues[] = {0, 0, 0, 0};
//byte rgb_values[]   = {0, 0, 0, 0};

#define NUM_OF_COLORS sizeof(rgb_pins)
int speedtable[] = { 0, 100, 50, 2 };
#define NUM_OF_SPEEDS sizeof(speedtable)

struct
{
  byte values[4];
  byte speedsetting;
  byte mode;
  bool status;
} rgb = { {0, 0, 0, 0}, 0, RAINBOW, false};

bool    flashOn      = true;
int     syscounter   = 0;
int     lastUpdate   = 0;
bool    newSetting   = false;


// MySensor
MyMessage msgMotor( 1 , V_STATUS);  // 0/1
MyMessage msgRGB( 2 , V_RGB );     // 0xRRGGBB
MyMessage var1( 0 , V_VAR1);       // Warning message


// Misc variables
boolean isMetric = true;
#define BAT_CHECK_INTERVAL 43200000;   // Check batteries every 12h 
unsigned long next_bat_check = 0; 
int batLevel = -1; 

void setup() {

  // Serial init
  Serial.begin( MY_BAUD_RATE );

  // Startup Info
  Serial.print("Board:");
  Serial.print( ARDUINO );
  Serial.print( "\t" );
  Serial.print( F_CPU / 1000000 );

  Serial.print(" Mhz\t");
  Serial.print( NAME );
  Serial.print(" ");
  Serial.println( RELEASE );

  Serial.print("Nrf24 CE/CS:");
  Serial.print( MY_RF24_CE_PIN);
  Serial.print("/");
  Serial.print( MY_RF24_CS_PIN );

  Serial.print("\tNODE ID:");
  Serial.print(hwReadConfig(EEPROM_NODE_ID_ADDRESS));

  isMetric = getControllerConfig().isMetric;
  Serial.print(F("\tisMetric:")); Serial.println(isMetric);

#ifdef OTA_WAIT_PERIOD
  Serial.print("OTA FW Enabled");
#endif


  // Fetchs previous state from controller
  request( 1 , V_RGB );
  //request( 2 , V_RGB );

  // Set the rgb(w) pins in output mode
  for (int i = 0; i < NUM_OF_COLORS; i++) {
    pinMode(rgb_pins[i], OUTPUT);
  }
  //recallEeprom();
  //setLedValues(rgb.values, true);

  Serial.println("ONLINE!");
  // Serial.flush();

}

void presentation() {

  delay(100); 
  sendSketchInfo( NAME , RELEASE );

  delay(100); 
  present( 1, S_RGBW_LIGHT );

}

void loop() {

  //Battery check 
  if ( millis() > next_bat_check ) { 
     
     batLevel = map( readVcc() , 2550 , 4200 , 0 , 100 ); 
     sendBatteryLevel( batLevel , false ); 
     next_bat_check += BAT_CHECK_INTERVAL ; 
  }
  
  if (speedtable[rgb.speedsetting] > 0) {
    if ((syscounter % speedtable[rgb.speedsetting]) == 0) {
      switch (rgb.mode)
      {
        case RAINBOW:
          animateRainbowStep();
          break;

        case FADERGB:
          animateFadeColorStep(FADE_RGB_COLOR_MODES);
          break;

        case FADEMULTICOLR:
          animateFadeColorStep(FADE_MULTI_COLOR_MODES);
          break;

        case FLASHCOLOR:
          setLedValues(flashOn ? ledOffValues : rgb.values, false);
          flashOn = !flashOn;
          break;

        case RANDONMIZE:
          long number = random(0, 0xFFFFFF);
          rgb.values[0] = number >> 16 & 0xFF ;
          rgb.values[1] = number >> 8 & 0xFF ;
          rgb.values[2] = number & 0xFF;
          setLedValues(rgb.values, false);
          break;
      }
    }
    delay(rgb.mode == RANDONMIZE || rgb.mode == FLASHCOLOR ? 50 : 1);
  }
  if (newSetting && (lastUpdate + 30000 < syscounter)) {
    // Wait for a couple of seconds be fore actual storing the current setting in to EEPROM
    // This will save the EEPROM's life time, when playing around with colors
    Serial.println(" Store EERPOM");
    // storeEeprom();
    newSetting = false;
  }
  delay(1);
  syscounter++;

}

void animateRainbowStep()
{
  static float counter = 0;
  float        pi      = 3.14159;
  counter++;
  rgb.values[0] = map(sin(counter / 100         ) * 1000, -1000, 1000, 0, 255);
  rgb.values[1] = map(sin(counter / 100 + pi * 2 / 3) * 1000, -1000, 1000, 0, 255);
  rgb.values[2] = map(sin(counter / 100 + pi * 4 / 3) * 1000, -1000, 1000, 0, 255);
  setLedValues(rgb.values, false);
}

void animateFadeColorStep(byte *modes)
{
  static int modecnt = 0;
  if (updateRGBValues(modes[modecnt] >> 1, (modes[modecnt] & 0x1) == 0x1)) {
    modecnt = (modes[modecnt + 1] == 0xFF ? 0 : modecnt + 1);
  }
}

bool updateRGBValues(byte mode, bool down)
{
  bool endReached = false;
  for (byte i = 0; i < 3; i++) {
    if (((mode >> i) & 0x1) == 0x1) {
      rgb.values[i] += (down ? -1 : 1);
      endReached    |= (down && (rgb.values[i] == 0x00)) || (!down && (rgb.values[i] == 0xFF));
    }
  }
  setLedValues(rgb.values, false);
  return endReached;
}

void receive(const MyMessage &message)
{
  if (message.type == V_RGB || message.type == V_RGBW) {
    // starting to process the hex code
    String hexstring = message.getString();

    long number;
#ifdef RGBW
    char white[3];
    white[0] = hexstring[6];
    white[1] = hexstring[7];
    white[2] = 0;
    number = (long) strtol( &white[0], NULL, 16);
    rgb.values[3] = number & 0xFF;
#endif
    hexstring[6] = 0;
    number = (long) strtol( &hexstring[0], NULL, 16);
    rgb.values[0] = number >> 16 & 0xFF ;
    rgb.values[1] = number >> 8 & 0xFF ;
    rgb.values[2] = number & 0xFF;

    rgb.speedsetting = 0;
    setLedValues(rgb.values, true);
    lastUpdate = syscounter;
    newSetting = true;
  }

  if (message.type == V_STATUS) {
    if (message.getBool()) {
      Serial.println("ON: Switch to last known color values");
      setLedValues(rgb.values, true);
    } else {
      Serial.println("OFF: Switch colors off");
      setLedValues(ledOffValues, true);
      next_bat_check = millis();  // Force battery level check 
    }
    rgb.speedsetting = 0;
    rgb.status = message.getBool();
    lastUpdate = syscounter;
    newSetting = true;
  }

  if (message.type == V_VAR1) {
    Serial.println("Set speed and program value");
    byte newsetting = message.getByte();
    rgb.speedsetting = (newsetting >> 4) & 0x0F;
    byte newmode = newsetting & 0x0F;

    if (newmode != rgb.mode) {
      for (byte i = 0; i < NUM_OF_COLORS; i++) {
        rgb.values[i] = 0;
      }
      rgb.mode = newmode;
    }
    if (rgb.speedsetting > 0) {
      rgb.status = true;
    }
    lastUpdate = syscounter;
    newSetting = true;

    Serial.print("Data 0x");
    Serial.print(newsetting, HEX);
    Serial.print(" speed:");
    Serial.print(rgb.speedsetting);
    Serial.print(" mode:");
    Serial.println(rgb.mode);
  }

}

void setLedValues(byte *rgb, bool show)
{
  for (int i = 0; i < NUM_OF_COLORS; i++) {
    analogWrite(rgb_pins[i], rgb[i]);
  }

  if (show) {
    Serial.print("Red: " );
    Serial.print(rgb[0], HEX);
    Serial.print("  Green: " );
    Serial.print(rgb[1], HEX);
    Serial.print("  Blue: " );
    Serial.print(rgb[2], HEX);
#ifdef RGBW
    Serial.print("  White is " );
    Serial.print(rgb[3], HEX);
#endif
    Serial.println();
  }
}


void storeEeprom()
{
  byte address = 0;
  byte *p = (byte *) & (rgb);
  for (byte i = 0; i < sizeof(rgb); i++) {
    saveState(address++, p[i]);
  }
}

void recallEeprom()
{
  byte address = 0;
  byte *p = (byte *) & (rgb);
  for (byte i = 0; i < sizeof(rgb); i++) {
    p[i] = loadState(address++);
  }
}

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA, ADSC));
  result = ADCL;
  result |= ADCH << 8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}
